/* groups.c: This program finds all groups of a given order n (3 or more),
   by looking for all the possible multiplication tables that represent
   a group, i.e. a latin square that also satisfies the associative law
   (ab)c=a(bc) for all elements a,b,c. A latin square is a square matrix
   of size nxn in which every row and column is a permutation of the
   integers 1,2,...,n. In this program, the multiplication table is
   represented by matrix elements t[i][j] (i=1 to n, j=1 to n), the
   group elements are represented by the integers 1,2,...,n (1 is
   the identity element), and t[i][j] is either 0 (if the product of
   elements i,j is not known yet) or the number representing the product
   (element i)(element j) by the group operation. Groups are tested
   for being isomorphic to a previous group.

   (c) Copyright by Dave Edwards, 2007. All rights reserved.

   NOTICE:
     The program is provided on an "as is" basis, without any support or
     warranty. If you decide to use it, you do so at your own risk.

   Notes:
   - Terminology used here: "Found" means any group (multiplication table)
     found by the search algorithm. "Processed" means a found group that
     is to be further worked on, such as to be displayed, saved, or
     tested for isomorphism with previously processed groups.
     For example, cyclic groups may be found, but are then skipped,
     and are not counted as processed. There are options to skip (i.e.
     not process) some types of groups.
   - The cyclic group (group with an element whose order is n) is found
     (many times for some n) but is skipped (not processed).
   - Two different multiplication tables found by this program may
     represent the same abstract group, since the 2 groups found may be
     isomorphic (by a reordering of the group symbols 1,2,...,n).
     There is an option (on by default) that determines whether each
     processed group is isomorphic to a previously processed group.
   - Since each element of the array t is an unsigned short int (16 bits),
     the maximum group order n supported is 2**16-1 = 65535.
     The actual maximum (normally much lass than 2**16-1) is defined by MAXN.
     The specific value of n (3 to MAXN) is defined by CURN.
     If CURN is defined as 0, the program prompts for n.
     In actual practice, execution time becomes prohibitively large except
     for quite small n. Before version 5, even n=16, 18, 27 were impractical;
     n=20 was doable but took several hours (see output in dir groups_out).
     However, sometimes at least 1 non-Abelian group can be found quickly,
     and the program is fast for n=pq (product of 2 different primes)
     and n=p**2 (square of a prime). Starting with version 5, a faster
     method is used for other n, and n=16, 18, 20, 27, 32, etc. are fast.
     With version 5, the only problem n <= 81 are n=60, 64.
     With version 6, n=64 is able to finish in several hours.
   - We do not use index values 0 in the array t, therefore each dimension
     is declared as MAXN+1 (=MAX1).
   - For larger values of MAXN and n, a very large stack size may be needed.
     The max number of nested calls to next() is approximately (n-1)**2,
     and each call allocates 2 tables, totalling about 4*((MAXN+1)**2) bytes
     per call = about 27 KB per call if MAXN is 81.
     If n is near MAXN, this is about 4*(MAXN**4) bytes!
     In actual practice, it seems that the number of nested calls to next()
     is much less than (n-1)**2, and seems to be less than n in the cases I
     have seen. Function don() is also recursive, but its nest level is
     low, the highest being about k for n=2**k.
     For the Mingw C compiler on Win32, the default stack seems to be
     about 1 MB. I could not find any documentation on Mingw stack size
     or how to increase it. If the stack is exceeded, it seems that the
     program just stops, with no error message.
   - We assume only the simple theorems of group theory: Lagrange's Theorem
     (the order of any subgroup is a divisor of the order of the group),
     my Theorem 4 (if all non-identity elements of a finite group G have
     order 2, then G is isomorphic to C2 x C2 x ... x C2), Product Theorem.
     An exception is that we assume Cauchy's Theorem (if prime p divides
     the order of G, then G has an element of order p) for some values of n.
     Also, for some values of n we assume the Sylow Theorems.
     If Cauchy or Sylow is needed, a message is issued.
   - References to Ledermann are to the book "Introduction to the Theory
     of Finite Groups" by W. Ledermann, 5th edition, 1964, published by
     Oliver and Boyd.
   - Job exit codes:
       0  Normal.
       2  Stopped when first non-Abelian group found.
       3  Logic error.
       4  Memory area (e.g. malloc() failed).
       5  Error in getgroup().
       8  Stopped when a particular group found (special coding).
       9  Stopped during debugging/testing.

   Method summary:

     Since the search for groups of order n may involve a preliminary
     search for groups of order n1 (where n1 is a factor of n, and n1<n),
     the actual search is done in a recursive function don().

     In the function don(), the search is done as follows:

     The recursive function next() fills in one cell of the table.
     (In some cases, the assocok() function may result in a call to
     next() filling in more than one cell.)
     However, next() works with a copy s of the table t passed to it, and
     calls found() when the table is complete (a group), therefore next()
     does not modify the table passed to it.

     The function assocok() checks that the associative law is satisfied.
     It can also modify the mult. table passed to it, filling in empty
     cells in order to satisfy the associative law.

     The function found() is called when any group is found. It determines
     characteristics of the group (whether abelian, number of elements
     of a given order, etc.), and may further process the group, depending
     on various options.

     The key to keeping the number of found groups to a reasonable size
     (to avoid impossibly long execution times) is to initialize the
     multiplication table as far as possible before starting the search.
     If possible, "assign" each of the elements of the group, so that
     the same abstract group, the same except for a reordering of the
     elements, is not found many times. Usually this means assuming the
     existence of 2 subsets (subgroups if possible) A (m1 elements)
     and B (m2 elements) such that n=m1*m2 and G=AB. Here, AB means the
     set of all products ab where a is in A and b is in B. Then the
     elements of a are assigned to elements 1 thru m1 of G, and the
     remaining elements of G are assigned to the remaining products in AB.
     This may require enumerating all the groups of order m1 and m2,
     and doing s1*s2 cases, where s1=# groups of order m1, s2=# groups of
     order m2. See routine don() for details. Things are easier if m1
     and/or m2 is a prime. Theorems that may be needed:
     - The Product Theorem: If A and B are subgroups and their intersection
       is only the identity element, then AB has n1*n2 elements, where
       n1,n2 are the orders of A,B.  Note that AB is a group only if AB=BA.
       If A or B is an invariant (=normal) subgroup of G then AB=BA, even
       if the other (B or A) is not a subgroup - see Ledermann p.100.
       The Product Theorem is similar in difficulty to Lagrange's Theorem,
       and is easily proved by decomposing B into cosets of the intersection
       of A and B, which is a subgroup.
     - Cauchy's Theorem (see above).
     - The Sylow Theorems: If p is a prime and p**t (t>0) divides the
       order n of a group G but p**(t+1) does not divide n, then G has
       at least 1 subgroup A of order p**t. A is the only such subgroup
       if and only if it is an invariant subgroup of G. If there are
       exactly k such subgroups, then p divides k-1 and k divides n.
       Also, for the case where n=p**t (p a prime and t>1), G has at
       least 1 invariant subgroup of order p**k, for all 0<k<t.

   Build procedure:

   groups.exe is created by MinGW C compiler (see c:\download\gcc\mingw),
   by the command (on Windows):
     gcc -Wall -W groups.c -o groups.exe
   Under Linux or Unix, compile by:
     gcc -Wall -W groups.c -o groups

Change Log:
18aug2007 - Written by Dave Edwards
31aug2007 - Version 3.
02sep2007 - Version 4: Add option to test whether each found group is
    isomorphic to previously found groups: new functions testisom(),
    nextisom() (recursive), isomok(). Table g_savedgr added.
  - In found(), use faster method of calculating the order of each
    element, save the orders in array ord[], and display the orders.
  - In savgroup(), also save ord[]. Add routine getgroup() to get a
    saved group.
  - Change chkisom(), including calling sequence.
  - Add timer coding (elapsed time).
08sep2007 - Version 5: Faster handling for n=pqs (product of 3 distinct
    primes, e.g. n=30). This uses the Sylow theorems (see my notes).
    Finds all the groups of that order quickly.
    Also for n=p**t and n=(p**t1)*(q**t2), where p and q are distinct
    primes and t>2 and (t1>1 or t2>1). Still very slow for n=60.
  - Add struct for saved group. Add idnum to savgroup(), getgroup().
    Group idnum is now saved as a separate item, instead of in ord[0].
  - Make search coding a recursive function don(), called from main
    and perhaps from itself. This allows search for groups of order n
    to enumerate groups of order n1 (where n1<n), in order to help
    in the search. Increase MAXSAV from 300 to 800, since each call
    to don() may save groups, cumulatively. Add g_firstsav.
  - Change needcauchy to g_needcauchy and add g_needsylow. These are
    now set in don(), as needed. The msgs are issued at end of main.
  - In don(), make the search method more general, using Sylow Theorems
    if needed. This makes search doable for n such as 16, 18, 20, 27, etc.
    n=60 is still a problem (impossibly slow). n=64 may be possible
    but takes many hours. All other n up to 81 are reasonably fast.
    n=81 takes about 31 minutes on a642.
  - In the case of n = 2**k, the group C2 x C2 x ... x C2 is now found.
  - Rename distrok() to assocok(), etc.
  - Exact meaning of g_debug changed in some cases.
  - Increase MAXN from 60 to 81.
23sep2007 - Version 6: Faster for n=p**t (power of a prime p, t>2).
    Add g_setfirsti, g_setfirstj, which may be set nonzero by don(),
    to tell next() to give priority to setting the product of (element i)
    and (element j). Result speed is about the same for n=3**t, but is
    faster for n=2**t. Much faster for n=32. For n=64, somewhat faster,
    but still very slow, finally finished in 46.6 hours on a642 machine
    (with version 5, was not complete when cancelled after 78 hours).
    Note: This change may cause groups to be found in a different order,
    which may cause intermediate group counts (e.g. when using g_debug=2)
    to be different, because of the order differences in lower n. But the
    final counts should be the same.
  - Better handling of n=60 and similar: use m2=pf[r-1] instead of 1.
    Now n=60 still does not finish (after 57.3 hours on a642), but at
    least it finds all except 1 non-Abelian group.
  - Change g_fndinterval from 500000 to 2000.
13oct2007 - Version 6.1: Use fgets and sscanf to read options, instead
    of scanf and fflush. The latter works fine on Win32, but not on Linux,
    since on Linux fflush(stdin) does nothing (but returns 0).
*/

#include <stdio.h>
#include <stdlib.h>   /* for exit(), malloc() */
#include <string.h>   /* for memcpy() */
#include <time.h>   /* for time() */

#define PROGVERSION "6.1"
#define MAXN 81   /* maximum n (3 to 2**16-1) */
#define MAX1 82   /* always MAXN+1 */
#define CURN 0    /* current value of n (3 to MAXN), = order of group.
                     0 means prompt for n. */

/* Structure for a saved group. See found() and savgroup() routines. */
struct savedgroupstruct
  { unsigned short int n;  /* order (# of elements) of the group */
    unsigned short int filler1;  /* not used */
    int idnum;  /* a nonzero id number assigned to this group (unique within
                   saved groups of this same n). Is the "nnn." in display. */
    unsigned short int t[MAX1][MAX1];  /* multiplication table: t[i][j]
                   (i,j=1 to n) is (element i)*(element j). Element 1 is
                   always the identity element. In group display, i is
                   (horizontal) row #, j is (vertical) column #. */
    unsigned short int chr[MAX1];  /* info about the group, see found().
                   This info is identical for isomorphic groups. */
    unsigned short int ord[MAX1];  /* orders of elements: ord[i] (i=1 to n)
                   is the order of element i */
  };

/* Global variables, local to this file */
/* NOTE: If variables are added here, you may also have to change the
     save and restore coding in don(). */
static int g_debug;   /* nonzero if debugging or testing (see comments below) */
static int g_skipabeliansw;  /* if nonzero, do not process abelian groups */
static int g_skipnonabelsw;  /* if nonzero, do not process non-abelian groups */
static int g_stopnonabelsw;  /* if nonzero, stop at 1st non-abelian group */
static int g_testisomsw=1;  /* if nonzero, test for isomorphisms between
                               processed groups */
static int g_displayisomsw=0;  /* if nonzero, display processed group even if
                             isom. to a previous processed group */
static int g_nodisplaysw;  /* if nonzero, skip all displays of
                              processed groups (but still count them) */
static int g_numgroupsf;  /* # of groups found */
static int g_numgroups;  /* # of groups processed */
static int g_numabelian;  /* # of abelian groups processed */
static int g_numnonabel;  /* # of non-abelian groups processed */
static int g_numcyclic;  /* # of cyclic groups found (they are not processed) */
static int g_numabelskip;  /* # of abelian non-cyclic groups found but
                                not processed */
static int g_numnonabelskip;  /* # of non-abelian groups found but
                                not processed */
static int g_numisomabelian;  /* # of processed abelian groups
                         isomorphic to a previous group (if g_testisonsw) */
static int g_numisomnonabel;  /* # of processed non-abelian groups
                         isomorphic to a previous group (if g_testisonsw) */
static int g_numnextcalls;  /* # of calls to next() */
static int g_nestlevel;  /* current next() call nest level */
static int g_hinestlevel;  /* highest next() call nest level */
static int g_numasokcalls;  /* # of calls to assocok() */
static int g_numasokcalls1;  /* # of calls to assocok() with return 1 */
static int g_numasokfilled;  /* # of cells filled by assocok() with return 1 */
static int g_numtestisomcalls;  /* # of calls to testisom() */
static int g_numnextisomcalls;  /* # of calls to nextisom() */
static int g_numisomokcalls;  /* # of calls to isomok() */
static int g_isomnestlevel;  /* current nextisom() nest level */
static int g_isomhinestlevel;  /* highest nextisom() nest level */
static void *g_savgrptr;  /* pointer to saved info for a particular
                             group: see savgroup() routine */
static int g_nsavedgr;  /* number of groups in g_savedgr table */
#define MAXSAV 800   /* max number of saved (nonisomorphic) groups */
static void *g_savedgr[MAXSAV];  /* array of pointers to info for
                  saved (nonisomorphic) groups: see savgroup() routine.
                  Used if testing whether each processed group is isomorphic
                  to any previously processed group. Note: # of groups of
                  order 32: 51, order 64: 267, order 96: 230, order 128: 2328,
                  order 256: 56092, therefore MAXSAV may need to be
                  increased if MAXN is increased. Also, if recursive calls
                  to don() are done for a particular n, this array may
                  need to hold groups found for lower values of n, in
                  addition to the current n. */
static int g_firstsav;  /* For a given group order n, this is the index
                  in g_savedgr[] of the first saved group of that order.
                  The index of the last saved group of that order is
                  g_nsavedgr-1. If g_firstsav >= g_nsavedgr, there are no
                  saved groups of that order. g_firstsav is valid in
                  next() and found() during the search for n (AFTER
                  don() has been called for lower n), and in a routine that
                  calls don() for n (immediately after the call to don()). */
static time_t g_timer1;  /* system time at start of search */
static time_t g_fndtime;  /* system timer (seconds) in found() routine */
static int g_fndtimecnt;  /* # of found groups at time g_fndtime */
static int g_fndinterval;  /* time interval in found(), as # of found groups */
static int g_needcauchy;  /* nonzero if Cauchy's Theorem was used for this
                  n (or by don() called for a lower value of n) */
static int g_needsylow;  /* nonzero if Sylow Theorems were used for this
                  n (or by don() called for a lower value of n) */
static int g_setfirsti,g_setfirstj;  /* priority cell for next() */

/* Prototypes for functions in this file */
static int don(unsigned short int n, int dotype);
static void next(unsigned short int t[MAX1][MAX1], unsigned short int n);
static int assocok(unsigned short int t[MAX1][MAX1], unsigned short int n);
static void found(unsigned short int t[MAX1][MAX1], unsigned short int n);
static void savgroup(unsigned short int n, int idnum,
     unsigned short int t[MAX1][MAX1], unsigned short int chr[MAX1],
     unsigned short int ord[MAX1], void **psavptr);
static void getgroup(unsigned short int n, int *pidnum,
     unsigned short int t[MAX1][MAX1], unsigned short int chr[MAX1],
     unsigned short int ord[MAX1], void *savptr);
static int chkisom(unsigned short int t[MAX1][MAX1],
     unsigned short int n, unsigned short int chrt[MAX1],
     unsigned short int ordt[MAX1], void *savptr);
static void factor(unsigned int n, unsigned int pf[], unsigned int pow[],
              unsigned int *r);
static int testisom(unsigned short int n,
          unsigned short int s[MAX1][MAX1], unsigned short int t[MAX1][MAX1],
          unsigned short int ords[MAX1], unsigned short int ordt[MAX1]);
static int nextisom(unsigned short int n, unsigned short int f[MAX1],
          unsigned short int s[MAX1][MAX1], unsigned short int t[MAX1][MAX1],
          unsigned short int ords[MAX1], unsigned short int ordt[MAX1]);
static int isomok(unsigned short int n, unsigned short int f[MAX1],
          unsigned short int s[MAX1][MAX1], unsigned short int t[MAX1][MAX1],
          int cntfset);

int main(void)
{
unsigned short int i,n;
int n1,n2,sw1,sizeans;
unsigned char ansch;
char ans[100];
unsigned int r;  /* number of distinct primes in factorization of n */
unsigned int pf[7],pow[7];  /* pf[i] (i=1 to r) are the distinct primes in the
           factorization of n, in increasing order. pow[i] is the number
           (1 or more) of times pf[i] divides n. pf[0] and pow[0] are not used.
           Since max n is 2**16-1 and 2*3*5*7*11*13*17 > that, we need
           space for only 6 primes, i.e. array size 7.
           E.g. n = 84 = 2**2 * 3 * 7: r=3 pf[1]=2 pow[1]=2 pf[2]=3 pow[2]=1
           pf[3]=7 pow[3]=1 .  Note: factor() is also called by don(),
           found(). */
time_t timer2;  /* system time in seconds */

printf("groups, version %s\n",PROGVERSION);
printf("(c) Copyright by Dave Edwards, 2007. All rights reserved.\n");
printf("Searches for all abstract groups of a given order (number of elements).\n");
printf("In group multiplication tables, xy is at row x column y.\n");

n=CURN;  /* # of elements in the group (3 to MAXN) */
sizeans=sizeof(ans);
for(;;)
  { if(n>=3 && n<=MAXN) break;
    printf("\nEnter group order n (3 to %d): ",MAXN);
    ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;  /* read next line */
    n1=0;
    n2=sscanf(ans,"%d",&n1);
    if(n2!=1) continue;   /* no value obtained */
    if(n1<3 || n1>MAXN) { printf("** Invalid, try again\n"); continue; }
    n=(unsigned short int)n1;
    break;
  }
printf("\nGroup order: n = %hd\n",n);

g_debug=0;   /* For debugging/testing: normal=0
                Values 1 to 10: extra info messages, with larger values
                   giving more info.
                   E.g. 1: minimal debugging messages.
                        2: display case (A,B) info and timing in don().
                        3: don(): display case info and initial tables.
                        5 or more: testisom() trace, including display of
                            isomorphism function f if isomorphic. Usually
                            option g_displayisomsw is also set.
                        6 or more: detailed tracing in isomok().
                Values 11 or more: to enable specific special coding.
                   E.g. 11: Detailed tracing in next() routine.
                        12: In found() routine, save a particular group,
                            compare it with a particular later group by
                            calling chkisom() routine.
                        44: Enable old method (version 4) coding in don().
             */
g_skipabeliansw=0;   /* init. */
g_skipnonabelsw=0;
g_stopnonabelsw=0;
g_testisomsw=1;
g_displayisomsw=0;
g_nodisplaysw=0;
g_numgroupsf=0;
g_numgroups=0;
g_numabelian=0;
g_numnonabel=0;
g_numcyclic=0;
g_numabelskip=0;
g_numnonabelskip=0;
g_numisomabelian=0;
g_numisomnonabel=0;
g_numnextcalls=0;
g_nestlevel=0;
g_hinestlevel=0;
g_numasokcalls=0;
g_numasokcalls1=0;
g_numasokfilled=0;
g_numtestisomcalls=0;
g_numnextisomcalls=0;
g_numisomokcalls=0;
g_isomnestlevel=0;
g_isomhinestlevel=0;
g_savgrptr=NULL;
g_nsavedgr=0;
g_firstsav=0;
g_fndinterval=2000;   /* # of found groups between time checks */
g_needcauchy=0;
g_needsylow=0;
g_setfirsti=0;
g_setfirstj=0;

/* Factor the group order, and report some theoretical results known
   for that order. Note: factor() is also called by don() and found(). */
factor((unsigned int)n,pf,pow,&r);  /* sets r, pf[i], pow[i] (i=1 to r) */
printf("   Factorization into prime divisors:\n   n = ");
for(i=1;i<=r;i++)
  { if(i>1) printf(" * ");
    printf("%u",pf[i]);
    if(pow[i]!=1) printf("**%u",pow[i]);
  }
printf("\n");

printf("Theoretical knowledge for this type of n:\n");
for(sw1=1,i=1;i<=r;i++) { if(pow[i]!=1) { sw1=0; break; } }
if(sw1)
  { /* Case 1: all pow[i] are 1 (n=product of 1 or more distinct primes). */
    if(r==1)
      { /* Case 1.1: n=prime. */
        printf("  n is a prime. The group is the cyclic group Cn.\n");
      }
    else if(r==2 && (pf[2]-1)%pf[1]!=0)
      { /* Case 1.2: n=pq where p and q primes, p<q, p does not divide q-1.
           By Ledermann Theorem 5 p.133 G is abelian, and by my Theorem 5
           G is in fact cyclic. E.g. n=3*5 */
        printf("  n=pq where p and q are primes, p<q, and p does not "
                     "divide q-1.\n");
        printf("  The group is the cyclic group Cn.\n");
      }
    else
      { /* Case 1.3: some other product of 2 or more distinct primes.
           By my Theorem 5 (for n=pq) or 6, if G is abelian it is cyclic. */
        printf("  n is a product of 2 or more distinct primes.\n");
        printf("  There may be non-Abelian groups, but any Abelian group\n"
               "  is the cyclic group Cn.\n");
      }
  }
else if(r==1 && pow[1]==2)
  { /* Case 2: n=p**2 (square of a prime).
       By my Theorem 7 (or Ledermann Theorem 5 p.104), G is abelian,
       and by my Theorem 5 G is Cn or Cp x Cp. */
    printf("  n is the square of a prime p.\n");
    printf("  Then the group is Abelian, and is either the cyclic group Cn\n"
           "  or the direct product Cp x Cp of 2 cyclic groups of order p.\n"
           "  Note: Cyclic groups will not be found in this case.\n");
  }
else printf("  (no simple results)\n");

/* Ask for the options */
printf("\n");
if(g_testisomsw) printf("Default is to test for isomorphisms.\n");
printf("Use standard options? ");
ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
ansch=0;
sscanf(ans," %c",&ansch);
if(ansch=='n' || ansch=='N')
  { printf("Enter debugging number (g_debug), normally 0: ");
    ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
    sscanf(ans," %d",&g_debug);
    if(g_debug) printf("Using g_debug = %d\n",g_debug);

    printf("Skip processing of Abelian (commutative) groups? ");
    ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
    ansch=0;
    sscanf(ans," %c",&ansch);
    if(ansch=='y' || ansch=='Y') g_skipabeliansw=1;

    printf("Skip processing of non-Abelian groups? ");
    ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
    ansch=0;
    sscanf(ans," %c",&ansch);
    if(ansch=='y' || ansch=='Y') g_skipnonabelsw=1;

    printf("Stop at first non-Abelian group found? ");
    ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
    ansch=0;
    sscanf(ans," %c",&ansch);
    if(ansch=='y' || ansch=='Y') g_stopnonabelsw=1;

    printf("Test for isomorphisms between processed groups? ");
    ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
    ansch=0;
    sscanf(ans," %c",&ansch);
    if(ansch!='n' && ansch!='N') g_testisomsw=1;
                            else g_testisomsw=0;  /* default is 1 */

    if(g_testisomsw)
      { printf("Display processed group even if isomorphic to a previous "
               "group? ");
        ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
        ansch=0;
        sscanf(ans," %c",&ansch);
        if(ansch=='y' || ansch=='Y') g_displayisomsw=1;

        printf("Suppress all group displays? ");
        ans[0]=0; fgets(ans,sizeans,stdin); ans[sizeans-1]=0;
        ansch=0;
        sscanf(ans," %c",&ansch);
        if(ansch=='y' || ansch=='Y')
          { g_nodisplaysw=1;
            g_displayisomsw=0;
          }
      }
  }

/* Pause, in case much output follows immediately */
printf("\nEnter any character to start search, or Ctrl+C to exit: ");
fgets(ans,sizeans,stdin);
printf("Searching...\n");

time(&g_timer1);  /* get start time in seconds */
g_fndtime=g_timer1;
g_fndtimecnt=0;

don(n,1);   /* do this n (as primary n) */

time(&timer2);  /* get end time in seconds */

printf("\nNumber of groups found of order %hu: %d\n",n,g_numgroupsf);
printf("Number of groups processed: %d\n",g_numgroups);
printf("Number of Abelian groups processed: %d\n",g_numabelian);
printf("Number of non-Abelian groups processed: %d\n",g_numnonabel);
printf("Number of cyclic groups found (not processed): %d\n",g_numcyclic);
if(g_skipabeliansw)
    printf("Number of Abelian non-cyclic groups found (not processed): %d\n",
           g_numabelskip);
if(g_skipnonabelsw)
    printf("Number of non-Abelian groups found (not processed): %d\n",
           g_numnonabelskip);
printf("Number of calls to next(): %d\n",g_numnextcalls);
printf("Highest next() call nest level: %d\n",g_hinestlevel);
if(g_debug)
  { printf("Number of calls to assocok(): %d\n",g_numasokcalls);
    printf("Number of calls to assocok() with return 1: %d\n",g_numasokcalls1);
    printf("Num of cells filled by assocok() with ret 1: %d\n",g_numasokfilled);
  }
if(g_testisomsw)
  { printf("Number of processed Abelian groups isomorphic to a previous group: "
           "%d\n",g_numisomabelian);
    printf("--> Number of distinct (non-isomorphic) processed Abelian groups: "
           "%d\n",g_numabelian-g_numisomabelian);
    printf("Number of processed non-Abelian groups isomorphic to a previous group: "
           "%d\n",g_numisomnonabel);
    printf("--> Number of distinct (non-isomorphic) processed non-Abelian groups: "
           "%d\n",g_numnonabel-g_numisomnonabel);
    printf("Num of calls to testisom(): %d\n",g_numtestisomcalls);
    printf("Num of calls to nextisom(): %d\n",g_numnextisomcalls);
    printf("Num of calls to isomok(): %d\n",g_numisomokcalls);
    printf("Highest nextisom() call nest level: %d\n",g_isomhinestlevel);
  }

if(r==1 && pow[1]==2) printf("--> NOTE: For this n (=p**2) cyclic groups "
              "are neither found nor processed.\n");
else printf("--> NOTE: Cyclic groups are found but not processed.\n");
if(g_needcauchy) printf("We used Cauchy's Theorem for this value of n.\n");
if(g_needsylow) printf("We used the Sylow Theorems for this value of n.\n");

n1=timer2-g_timer1;   /* elapsed time in seconds */
printf("Total elapsed time: %d seconds = %d minutes\n",n1,n1/60);

return 0;

}  /* end of main */

/* ---------------------------------------------------------------------- */

static int don(unsigned short int n, int dotype)
/* This is a recursive function that does the search for groups of
   order n. When called from main(), dotype is 1. If the search for
   groups of order n needs to enumerate groups of a lower order n1
   (a factor of n), don() calls don() for n1, with dotype=0.
   dotype=0 indicates that standard options should be used (i.e. test
   for isomorphic found groups, save distinct groups in g_savedgr[],
   etc.) and most messages suppressed. The index of the first group
   saved to g_savedgr[] for the specified n is available to the caller
   as g_firstsav. The caller sees the saved groups of order n (if any) as
   g_savedgr[i], i=g_firstsav to g_nsavedgr-1 inclusive.
   Note that the cyclic group Cn is found but not processed, therefore
   NOT saved.
*/
{
unsigned short int i,j,m1,m2,k,p,i2,j2;
int gi,gj,sava,savb,bnotsubgr,idnum;
int first1,lim1,first2,lim2;
time_t tim;
int sv[18];  /* for saving options and counts when dotype is 0 */
unsigned int r;  /* number of distinct primes in factorization of n */
unsigned int pf[7],pow[7];  /* pf[i] (i=1 to r) are the distinct primes in the
           factorization of n, in increasing order. pow[i] is the number
           (1 or more) of times pf[i] divides n. */
unsigned short int t[MAX1][MAX1];  /* group multiplication table */
unsigned short int s[MAX1][MAX1];  /* saved group */
unsigned short int chrs[MAX1],ords[MAX1];  /* saved group */

if(g_debug) printf("*Entering don() for n=%hu dotype=%d\n",n,dotype);

if(n<3)   /* degenerate case: only cyclic group */
  { g_firstsav=g_nsavedgr;  /* index of first saved group of order n */
    if(g_debug) printf("*Returning from don() for n=%hu, no groups saved\n",n);
    return 0;
  }

/* factor n into a product of powers of primes */
factor((unsigned int)n,pf,pow,&r);  /* sets r, pf[i], pow[i] (i=1 to r) */

if(dotype==0)
  { /* Save some existing options and counts */
    sv[0]=g_debug;
    sv[1]=g_skipabeliansw;
    sv[2]=g_skipnonabelsw;
    sv[3]=g_stopnonabelsw;
    sv[4]=g_testisomsw;
    sv[5]=g_displayisomsw;
    sv[6]=g_nodisplaysw;
    sv[7]=g_numgroupsf;
    sv[8]=g_numgroups;
    sv[9]=g_numabelian;
    sv[10]=g_numnonabel;
    sv[11]=g_numcyclic;
    sv[12]=g_numabelskip;
    sv[13]=g_numnonabelskip;
    sv[14]=g_numisomabelian;
    sv[15]=g_numisomnonabel;
    sv[16]=g_fndtimecnt;
    sv[17]=g_fndinterval;

    /* Reset some */
    if(g_debug>1) g_debug=1;  /* suppress detailed debugging */
    g_skipabeliansw=0;
    g_skipnonabelsw=0;
    g_stopnonabelsw=0;
    g_testisomsw=1;
    g_displayisomsw=0;
    g_nodisplaysw=1;         /* do not display groups */
    g_numgroupsf=0;
    g_numgroups=0;
    g_numabelian=0;
    g_numnonabel=0;
    g_numcyclic=0;
    g_numabelskip=0;
    g_numnonabelskip=0;
    g_numisomabelian=0;
    g_numisomnonabel=0;
    g_fndtimecnt=0;
    g_fndinterval=-1;     /* this suppresses progress msg coding in found() */
  }

/* Based on the factorization of n, determine how the mult. table is
   to be initialized for searching for groups.
   Use 2 subsets A and B of the group G. A is usually a subgroup.
   Order of A is m1. Order of B is m2. Order of AB (all products ab of
   elements a in A and b in B) is m1*m2. A is a subgroup, and AB is
   made up of m2 cosets of A. Note that m1*m2 may be less than n.
   Set: m1, m2 (m2=1 means no B, or B consists only of the identity element),
        first1, lim1 (relating to m1),
        first2, lim2 (relating to m2)
   The cases for A are the cyclic group of order m1 and (if first1<lim1)
   the saved groups g_savedgr[i] (i=first1 to lim1-1) resulting from call
   to don(m1,0). If first1=0=lim1, the cyclic group is the only case for A.
   Similarly for the cases for B (using m2, first2, lim2).
   Total # of cases (number of searches) is (1+lim1-first1)*(1+lim2-first2).
   Note: For some n (namely n=p**t where p is a prime and t>2), B is
   not a subgroup, but the m2 cosets Ab (where b is in B) are disjoint;
   switch bnotsubgr is set to 1 in that case, also indicating that B
   consists of the powers b**i (i=0 to m2-1) of some element b not in A.
*/
m2=1; first1=0; lim1=0; first2=0; lim2=0; bnotsubgr=0;  /* init. */
p=(unsigned short int)pf[1];
if(r==1)
  { if(pow[1]==1)
      { /* n = prime */
        m1=p;
      }
    else if(pow[1]==2)
      { /* n = square of a prime */
        /* Note: In this case, we assume G is not cyclic, therefore the
           order of b (an element not in A) is p (since it can't be p**2),
           therefore B is a subgroup. But this means that the cyclic
           group G will NOT be found by the search. main() issues a
           message to account for this. Note that, regardless of the type
           of n, cyclic groups are never saved. */
        m1=p;
        m2=p;
      }
    else
      { /* n = cube or higher power of a prime */
        m1=n/p;
        m2=p;
        bnotsubgr=1;  /* B is not necessarily a subgroup, since the order
               of b (an element not in A) may be > p, therefore b**p is
               not necessarily in B. (But we know b**p is in A.)
               The cosets A, Ab, A(b**2), ..., A(b**(p-1)) are disjoint
               and represent all the elements of G. B consists of the
               powers b**i (i=0 to m2-1). */
        g_needsylow=1;  /* need Sylow Theorems */
        don(m1,0);  /* get the groups A of order m1 = p**(pow[1]-1) */
        first1=g_firstsav;  /* the saved groups are the noncyclic of order m1 */
        lim1=g_nsavedgr;
      }
  }
else if(r==2)
  { /* n = p**pow[1] * q**pow[2] where p,q are primes and p<q */
    for(m2=p,gi=2;gi<=(int)pow[1];gi++) m2*=p;  /* set m2=p**pow[1] */
    m1=n/m2;  /* set m1=q**pow[2] */
    if(pow[2]>1)
      { don(m1,0);
        first1=g_firstsav;
        lim1=g_nsavedgr;
      }
    if(pow[1]>1)
      { don(m2,0);
        first2=g_firstsav;
        lim2=g_nsavedgr;
      }
    if(pow[1]==1 && pow[2]==1)
      { if(p!=2) g_needcauchy=1;  /* need Cauchy Theorem, except if n=2q */
      }
    else g_needsylow=1;
  }

else if(r==3 && pow[1]==1 && pow[2]==1 && pow[3]==1)
  { /* n = pqs where p,q,s are primes and p<q<s.
       It can be proved (from the Sylow theorems, by counting the number
       of elements of orders q and s - see my notes) that there is an
       invariant (normal) subgroup of G of order q or s.
       Therefore there are subgroups Q and S of order q and s, at least
       one of which is invariant, therefore QS is a subgroup of order qs.
       We use A = QS. */
    m2=p;
    m1=n/p;
    don(m1,0);
    first1=g_firstsav;
    lim1=g_nsavedgr;
    g_needsylow=1;
  }
else
  { /* Other n: I cannot think of a good (fast) strategy here.
       n is divisible by 4 or more distinct primes, or by 3 distinct
       primes but 1 or more of the 3 divides n more than once.
       (We cannot guaranty the existence of an invariant subgroup,
       even if r=3, since e.g. there is a simple group of order 60 (it
       has no invariant proper subgroups).)
       For now, use m1=highest prime divisor, m2=2nd-to-highest prime.
          A consists of powers of some element of order m1.
          B consists of powers of some element b (not in A) or order m2.
          Note that m1*m2 < n. A and B are subgroups but AB is not
          necessarily a subgroup (it is a subgroup if A or B is invariant
          or if G is Abelian).
       [A better method may be to decompose G relative to 2 subgroups
       A and B, where the orders of A and B are relatively prime, using
       Theorem 8 (Frobenius) in Ledermann p.60. Then G = A(c1)B + A(c2)B
       + ... + A(cm)B, where c1=1. Each (ci**(-1))A(ci) has the same order
       as A (in fact is isomorphic to A), therefore its intersection with B
       is 1. Therefore n=m1*m2*m where m1 is the order of A, m2 is the
       order of B, and all the elements of G can be assigned. E.g. for
       n=60, use m1=5, m2=3, m=4.]
       The search will take an excessive amount of time.
       The only values of n <= 100 for this case are 60 (=4*3*5),
       84 (=4*3*7), 90 (=2*9*5). */
    m1=(unsigned short int)pf[r];  /* use m1 = largest prime factor */
    m2=(unsigned short int)pf[r-1];  /* m2 = next lower prime factor */
    g_needcauchy=1;
  }

if(g_debug)
  { printf("  Initialization of multiplication table will use m1=%hu m2=%hu\n",
           m1,m2);
    printf("  first1=%d lim1=%d first2=%d lim2=%d bnotsubgr=%d\n",
           first1,lim1,first2,lim2,bnotsubgr);
  }

g_firstsav=g_nsavedgr;  /* will be index of first saved group of order n */

/* Loops to do search for each pair A,B.
   gi=0 is for A = cyclic group of order m1.
   gi=1,2,... is for A = saved groups (if any) of order m1.
   Similarly for gj for B (if m2 is not 1). */
for(gi=0;;gi++)
  { if(gi>0)
      { sava=first1+gi-1;
        if(sava>=lim1) break;
      }
    for(gj=0;;gj++)
      { if(gj>0)
          { savb=first2+gj-1;
            if(savb>=lim2 || m2==1) break;
          }
        if(g_debug>=2 && dotype==1)
          { time(&tim);
            printf("------- Starting search case "
                  "gi=%d gj=%d  So far: %d+%d fnd=%d @ %lu secs\n",gi,gj,
                  g_numabelian-g_numisomabelian,
                  g_numnonabel-g_numisomnonabel,g_numgroupsf,tim-g_timer1);
          }

        /* Init. the multiplication table t: */
        g_setfirsti=0; g_setfirstj=0;  /* init. (priority cell for next()) */
        if(g_debug>=3 && dotype==1)
          { /* Init. full table, for check below */
            for(i=0;i<MAX1;i++) for(j=0;j<MAX1;j++) t[i][j]=9999;
          }
        for(i=1;i<=n;i++) for(j=1;j<=n;j++) t[i][j]=0;
        for(i=1;i<=n;i++)
          { t[1][i]=i;  /* column 1 */
            t[i][1]=i;  /* row 1 */
          }

        /* Init. t for A */
        if(gi==0)
          { /* A is cyclic group of order m1 */
            /* Set all products (elem i)(elem j) where i and j are 2 to m1 */
            for(i=2;i<=m1;i++) for(j=2;j<=m1;j++)
              { k=(i+j-2)%m1;
                t[i][j]=k+1;
              }
          }
        else
          { /* A is a saved group of order m1 */
            getgroup(m1,&idnum,s,chrs,ords,g_savedgr[sava]);  /* sets s etc. */
            for(i=2;i<=m1;i++) for(j=2;j<=m1;j++) t[i][j]=s[i][j];
          }

        /* Init. t for B (if any) */
        /* Elements of A: a1=1, a2, ..., am1.
           Elements of B: b1=1, b2, ..., bm2. */
        /* First, assign elements m1+1, m1+2, ..., 2*m1 = coset Ab2,
           elements 2*m1+1, 2*m1+2, ..., 3*m1 = coset Ab3, ... coset Abm2.
           (If m2=1 this does nothing.) */
        for(i=2;i<=m1;i++) for(j=m1+1;j<=m1*m2;j+=m1) t[i][j]=j+i-1;
        /* Next, if B is not a subgroup, set t to indicate that
           b3=b2**2, b4=b2**3(=b2*b3=b3*b2), ..., bm2=b2**(m2-1)(=...).
           (If m2=1 this does nothing.)
           Else, if B is a subgroup, set all bi*bj from saved group, or
           as cyclic group B if gj=0. (Does nothing if m2=1.) */
        if(bnotsubgr)
          { for(i=m1+1,j=m1+1;j<=m1*(m2-1);j+=m1)  /* does nothing if m2<=2 */
              { t[i][j]=j+m1;
                t[j][i]=j+m1;
              }
            /* Also, set g_setfirsti and g_setfirstj to tell next()
               routine to give priority to setting cell t[i][j].
               This is meant to make processing faster, for n=p**t
               where p (=m2) is a prime and t>2. See my theoretical
               notes of 22sep2007.
               We specify cell t[i][j] for the product (b2)**(m2) =
               (b2)(b2**(m2-1)) = (b2)(bm2). This product is in A, but
               it is not necessarily 1, since B is not a subgroup.
               Setting that cell early in next() makes the search faster.
               E.g. If p is 2 (n=2**t), then i=m1+1, j=m1+1, i.e. we give
               priority to setting b2**2 (=b**2). */
            if(m2>1)
              { g_setfirsti=m1+1;
                g_setfirstj=m1*(m2-1)+1;
                if(g_debug>=2 && dotype==1) printf("        Using "
                    "g_setfirsti=%d g_setfirstj=%d\n",g_setfirsti,g_setfirstj);
              }
          }
        else
          { if(gj==0)
              { /* B is cyclic subgroup of order m2 (powers of b2) */
                for(i=2;i<=m2;i++)
                  { i2=1+m1*(i-1);
                    for(j=2;j<=m2;j++)
                      { j2=1+m1*(j-1);
                        k=(i+j-2)%m2;
                        t[i2][j2]=1+m1*k;
                      }
                  }
              }
            else
              { /* Get B as saved group of order m2 */
                getgroup(m2,&idnum,s,chrs,ords,g_savedgr[savb]);
                for(i=2;i<=m2;i++)
                  { i2=1+m1*(i-1);
                    for(j=2;j<=m2;j++)
                      { j2=1+m1*(j-1);
                        k=s[i][j];
                        t[i2][j2]=1+m1*(k-1);
                      }
                  }
              }
          }

        if(g_debug>=3 && dotype==1)
          { printf("   Initial table before call assocok():\n");
            for(i=1;i<=n;i++)
              { printf("       ");
                for(j=1;j<=n;j++) printf(" %3hu",t[i][j]);
                printf("\n");
              }
            /* Also check that we stayed within n x n */
            for(i=0;i<MAX1;i++) for(j=0;j<MAX1;j++)
              { if((i==0 || i>n || j==0 || j>n) && t[i][j]!=9999)
                  { printf("**** Logic error: element t[%hu][%hu] "
                           " = %hu was set.\n",i,j,t[i][j]);
                    exit(3);
                  }
              }
          }

        /* Call assocok() to fill in additional products in the table,
           implied by the above. If assocok() finds an inconsistency,
           then this case (A,B) is not possible. */
        if(assocok(t,n)==0)
          { if(g_debug>=2 && dotype==1) printf("   Rejected by assocok()\n");
            continue;  /* skip to next case */
          }

        if(g_debug>=3 && dotype==1)
          { char ans[100];

            printf("   Initial table:\n");
            for(i=1;i<=n;i++)
              { printf("       ");
                for(j=1;j<=n;j++) printf(" %3hu",t[i][j]);
                printf("\n");
              }
            /* Pause before continuing. This disturbs timing, but we are
               only debuging. */
            printf("Enter any character to continue: ");
            fgets(ans,sizeof(ans),stdin);
          }

        next(t,n);  /* start the search, for this case (A,B) */

        if(g_nestlevel!=0) printf("*** Warning: g_nestlevel = %d should be 0 at end\n",
                              g_nestlevel);

      }  /* end of "for gj" */
  }  /* end of "for gi" */

if(dotype==0)
  { /* Restore options and counts */
    g_debug=sv[0];
    g_skipabeliansw=sv[1];
    g_skipnonabelsw=sv[2];
    g_stopnonabelsw=sv[3];
    g_testisomsw=sv[4];
    g_displayisomsw=sv[5];
    g_nodisplaysw=sv[6];
    g_numgroupsf=sv[7];
    g_numgroups=sv[8];
    g_numabelian=sv[9];
    g_numnonabel=sv[10];
    g_numcyclic=sv[11];
    g_numabelskip=sv[12];
    g_numnonabelskip=sv[13];
    g_numisomabelian=sv[14];
    g_numisomnonabel=sv[15];
    g_fndtimecnt=sv[16];
    g_fndinterval=sv[17];
  }

if(g_debug) printf("*Returning from don() for n=%hu, %d groups saved\n",
                   n,g_nsavedgr-g_firstsav);
return 0;
}  /* end of don */

/* ---------------------------------------------------------------------- */

static void next(unsigned short int t[MAX1][MAX1], unsigned short int n)
{
unsigned short int i,j,k,i1,j1,p1,ii,jj;
unsigned int nposs,lowposs;
unsigned short int s[MAX1][MAX1],s2[MAX1][MAX1];
unsigned char used[MAX1];  /* element value 1=used, 0=ok for current cell */

g_numnextcalls++;
g_nestlevel++;  /* decrement before return */
if(g_nestlevel>g_hinestlevel) g_hinestlevel=g_nestlevel;
if(g_debug==11)   /* special coding: */
  { printf("next() called at level %d\n",g_nestlevel);
    printf(" t =   ");  /* display the mult. table */
    for(i=1;i<=n;i++)
      { for(j=1;j<=n;j++) printf(" %3hu",t[i][j]);
        printf("\n       ");
      }
    if(g_nestlevel>10) { printf("*** stopping\n"); exit(9); }
  }

/* Make a working copy of the table */
for(i=1;i<=n;i++) for(j=1;j<=n;j++) s[i][j]=t[i][j];

for(;;)  /* main loop: lets us continue after filling in a cell that
            has only 1 possibility for its value. This saves stack space. */
  { /* Find an empty cell with fewest possible values.
       Possible values for a cell are the element numbers not already
       occurring in the cell's row and column.
       Exception: if g_setfirsti is nonzero, give priority to setting
       cell i1=g_setfirsti, j1=g_setfirstj.  See don(). */
    i1=0; j1=0;  /* if nonzero: cell with fewest possible values so far */
    lowposs=0;  /* none yet */
    for(i=2;i<=n;i++)
      { if(lowposs==1) break;
        for(j=2;j<=n;j++)
          { if(s[i][j]) continue;
            for(k=1;k<=n;k++) used[k]=0;  /* init. */
            for(k=1;k<=n;k++)
              { p1=s[i][k]; if(p1) used[p1]=1;
                p1=s[k][j]; if(p1) used[p1]=1;
              }
            for(nposs=0,k=1;k<=n;k++) if(used[k]==0) nposs++;  /* count # of possible values */
            if(nposs==0) { g_nestlevel--; return; }  /* group not possible */
            if(lowposs==0 || nposs<lowposs)
              { lowposs=nposs;
                i1=i; j1=j;  /* remember location of cell */
                if(nposs==1) break;  /* if only 1 possibility, choose that cell */
                                /* Break out of i loop also */
              }
          }
      }
    if(g_setfirsti)   /* Cell priority specification by caller */
      { /* Note: we still need the above loop, in order to set lowposs.
           Also, if we found a cell with only 1 possibility, use it! */
        if(g_setfirstj!=0 && s[g_setfirsti][g_setfirstj]==0 && lowposs>1)
          { i1=(unsigned short int)g_setfirsti;
            j1=(unsigned short int)g_setfirstj;
          }
      }
    if(i1==0)
      { /* All cells filled in: we have a group */
        found(s,n);
        g_nestlevel--; return;
      }
    /* If only 1 possibility for the cell value, fill it in, then iterate.
       Whenever we fill in a cell, if value is 1 (the group's identity
       element), also set the corresponding cell on the other side of the
       diagonal to 1, since xy=1 implies yx=1. Also call assocok().
       assocok() does the same when setting a cell to 1.
       Note: for lowposs==1, we already have the used[] array. */
    if(lowposs==1)
      { for(p1=1;p1<=n;p1++) if(used[p1]==0) break;  /* find cell value */
        s[i1][j1]=p1;  /* complete the cell */
        if(p1==1 && i1!=j1)
          { if(s[j1][i1]) { g_nestlevel--; return; }  /* not a group */
            s[j1][i1]=p1;
          }
        if(!assocok(s,n)) { g_nestlevel--; return; }  /* not a group */
        continue;  /* iterate main loop */
      }
    /* There are 2 or more possible values for the cell (i1,j1).
       Reget the used[] array, then call next() for each of the
       possible values.  See notes about cell value 1 above.
       Caution: assocok() may fill in additional values, therefore
       we must copy array s to another working copy s2, so that we
       can get back to the original s for each try. */
    i=i1; j=j1;
    for(k=1;k<=n;k++) used[k]=0;  /* init. */
    for(k=1;k<=n;k++)
      { p1=s[i][k]; if(p1) used[p1]=1;
        p1=s[k][j]; if(p1) used[p1]=1;
      }
    for(p1=1;p1<=n;p1++)  /* Loop thru the possible values p1 */
      { if(used[p1]) continue;
        for(ii=1;ii<=n;ii++) for(jj=1;jj<=n;jj++) s2[ii][jj]=s[ii][jj]; /* copy */
        s2[i][j]=p1;  /* assign the cell */
        if(p1==1 && i!=j)
          { if(s2[j][i]) continue;  /* this group is not possible */
            s2[j][i]=p1;
          }
        if(!assocok(s2,n)) continue;  /* no group */
        next(s2,n);  /* explore this possibility */
      }
    break;  /* exit from main loop */
  }  /* end of main loop */

g_nestlevel--;
}  /* end of next */

/* ---------------------------------------------------------------------- */

static int assocok(unsigned short int t[MAX1][MAX1], unsigned short int n)
/* The assocok function checks that all products defined so far in
   mult. table t satisfy the associative law, and returns 1 if yes,
   0 if no. Note: If any of i,j,k is 1, (ij)k is automatically = i(jk),
   therefore we can skip the check.
   Special case:
   If both ij and jk are defined, but one of (ij)k,i(jk) is defined and
   the other not, we can fill in an empty cell (provided the product
   does not already appear in the cell's row or column). In that case,
   we modify t. If the product already appears in the cell's row or
   column, we return 0.
   Also, when filling in a product xy like this, if xy=1 and x and y
   are different, we also fill in yx (=xy), since x and y are inverses
   of each other.
   Note: If return 0, any changes may to t should be discarded by the
   caller, i.e. the caller should recopy the table (except when assocok()
   is called by main() when initializing the table).
*/
{
unsigned short int i,j,k,ij,jk,p1,p2,r;
int setsw,numfilled;

g_numasokcalls++;
numfilled=0;
for(;;)
  { setsw=0;  /* will be set to 1 if an empty cell is filled in */
    for(i=2;i<=n;i++) for(j=2;j<=n;j++)
      { ij=t[i][j];  /* product of elems i and j */
        if(ij)  /* can skip check if ij is 0 (not defined yet) */
          { for(k=2;k<=n;k++)
              { jk=t[j][k]; if(jk==0) continue;  /* jk not defined */
                p1=t[ij][k];   /* (ij)k */
                p2=t[i][jk];   /* i(jk) */
                if(p1==0)
                  { if(p2)
                      { /* Perhaps we can set t[ij][k]=p2 */
                        for(r=1;r<=n;r++)
                          { if(t[ij][r]==p2 || t[r][k]==p2) return 0;
                          }
                        t[ij][k]=p2;
                        setsw=1;
                        numfilled++;
                        /* If we are filling in xy, and xy=1 and x and y are different,
                           then they are inverses of each other, therefore also fill
                           in yx with the same value as xy. */
                        if(p2==1 && ij!=k)
                          { if(t[k][ij]) return 0;
                            t[k][ij]=p2;
                            numfilled++;
                          }
                      }
                  }
                else
                  { if(p2)
                      { if(p1!=p2) return 0;
                      }
                    else
                      { /* Perhaps we can set t[i][jk]=p1 */
                        for(r=1;r<=n;r++)
                          { if(t[i][r]==p1 || t[r][jk]==p1) return 0;
                          }
                        t[i][jk]=p1;
                        setsw=1;
                        numfilled++;
                        if(p1==1 && i!=jk)  /* handle inverses, as above */
                          { if(t[jk][i]) return 0;
                            t[jk][i]=p1;
                            numfilled++;
                          }
                      }
                  }
              }
          }
      }
    if(setsw==0) break;  /* if we filled an empty cell, check assoc. again */
  }
g_numasokcalls1++;
g_numasokfilled+=numfilled;
return 1;
}  /* end of assocok */

/* ---------------------------------------------------------------------- */

static void found(unsigned short int t[MAX1][MAX1], unsigned short int n)
/* A group of order n has been found. Decide whether to process it.
   If so, get info about the group and its elements, test for an
   isomorphism with a previous group and save the group if new (if
   requested), and perhaps display the group.
*/
{
unsigned short int i,j,k,t1,p1,p2;
int sw1;  /* temp. switch */
int displsw;  /* local display switch */
int groupidnum,previdnum;  /* group id numbers for save and display */
int newgrsw,gi,n1,n2;
void *savptr;  /* pointer to memory area for saved group */
unsigned short int chr[MAX1];  /* Characteristics of the group, the
       same for any 2 isomorphic groups: chr[1]=0 if abelian (commutative),
       1 if non-abelian. chr[i] = # of elements with order i (i=2,3,...,n).
       Note: The group's id number is in chr[0] when the group is
       saved. chr[0] is not part of the group's "characteristics". */
unsigned short int ord[MAX1];  /* order of each element in the group:
       ord[i] = order of element i (i=1 to n) */
unsigned short int powtmp[MAX1];  /* temp array used when calculating the
                        orders of elements */
unsigned int r,pf[7],powf[7];  /* used in factoring k */
unsigned short int s[MAX1][MAX1],chrs[MAX1],ords[MAX1];  /* saved group */
time_t timernow;

g_numgroupsf++;

if(g_numgroupsf-g_fndtimecnt==g_fndinterval)
  { time(&timernow);  /* get system time in seconds */
    n1=timernow-g_fndtime;  /* # of secs since last check */
    n2=timernow-g_timer1;  /* # of secs since start */
    if(n1>0) printf("Elapsed: %d secs = %d mins. Groups found: %d (%d+%d)"
                  " Rate/min: %0.1lf\n",
                  n2,n2/60,g_numgroupsf,
                  g_numabelian-g_numisomabelian,
                  g_numnonabel-g_numisomnonabel,
                  g_fndinterval/(n1/60.));
    g_fndtime=timernow;
    g_fndtimecnt=g_numgroupsf;
  }

/* Get the group's characteristics */
for(i=1;i<=n;i++) chr[i]=0;  /* init. */
for(sw1=0,i=2;i<=n;i++)    /* test whether abelian */
  { if(sw1) break;
    for(j=i+1;j<=n;j++)
      { if(t[i][j]!=t[j][i]) {chr[1]=1; sw1=1; break; }
                        /* Break out of i loop also */
      }
  }

/* if(chr[1]==0) return; */  /* **1** skip it if abelian */
              /* Comment **n** (n a number) indicates special coding which
                 is usually commented out */

/* Calculate the order ord[i] of each element i.
   We use a fast method, using the following facts:
   (1) An element and its inverse have the same order:
       If element a has order k, then a**(k-1) has same order as a,
       and a**i and a**(k-i) have the same order.
   (2) If an element a has order k, and (i,k)=1 i.e. i and k are
       relatively prime, then a**i has the same order as a.
   (3) As a special case of (2), if a has order p (a prime), then
       a**i (i=2 to p-1) have same order as a.
   (4) If an element a has order k, and gcd(a,k)=d, then a**i has order k/d.
       (This is used for particular small values of k.)
   Notes:
   - If A is the set of elements whose orders have already been calculated,
     and b is an element not in A, let b have order k and let B be the
     subgroup (of k elements) generated by b.  Let D be the intersection
     of sets A and B, and let D have d elements. Then:
   - If k is a prime (k>2), b**i (i=2 to k-1) each has order k (since (i,k)=1)
     and therefore b is a power of b**i. Therefore, assuming we set the
     orders of powers of a when we set order of a, then none of b**i
     (i=2 to k-1) has its order set yet.
   - Similarly b**(k-1) (for any k>2) has not had its order set yet,
     assuming we set the order of the inverse of a when setting the order
     of any a.
   - If A is a subgroup (e.g. after calculating the order of element 2
     and its powers), then D is a subgroup of B, therefore d divides k.
     But also d<k (since b is not in A). Letting q be the smallest prime
     that divides k and letting k=dx (where x is some integer > 1),
     we have x >= q (since if x<q, then any prime divisor of x would
     divide k and be less than q). Therefore the number of powers of b
     whose orders have not already been calculated is k-d >= k(1-1/q).
     (In the worst case (q=2), about 1/2 of the powers of b have not had
     their orders calculated yet.)
   - The above remarks indicate that the coding method used here is
     worthwhile.
*/
ord[1]=1;
for(j=2;j<=n;j++) ord[j]=0;  /* init. */
for(j=2;j<=n;j++)
  { if(ord[j]) continue;  /* already done */
    /* Calculate the order k of b=(element j) and (most of) its powers.
       Temporarily save the element # of b**i in powtmp[i] (i=2 to k-1). */
    for(t1=j,i=2;;i++)
      { t1=t[t1][j];  /* next power of b: b**i */
        if(t1==1) break;
        powtmp[i]=t1;
      }
    if(i==n)  /* skip the group now if it is cyclic */
      { g_numcyclic++;
        return;
      }
    ord[j]=i;  /* order of b is i */
    if(i==2) continue;  /* done if order <= 2 */
    k=i;
    ord[powtmp[k-1]]=k;  /* set the order of b**(k-1) (the inverse of b) */
    switch (k)
      { case 3: break;
        case 4: ord[powtmp[2]]=2;  /* order of b**2 is k/2 */
                break;
        case 5: ord[powtmp[2]]=k;  /* k is prime */
                ord[powtmp[3]]=k;
                break;
        case 6: ord[powtmp[2]]=3;
                ord[powtmp[3]]=2;
                ord[powtmp[4]]=3;
                break;
        case 8: ord[powtmp[2]]=4;
                ord[powtmp[3]]=8;
                ord[powtmp[4]]=2;
                ord[powtmp[5]]=8;
                ord[powtmp[6]]=4;
                break;
        default:
            factor((unsigned int)k,pf,powf,&r);
                      /* factor k: sets r, pf[i], powf[i] (i=1 to r) */
            if(r==1 && powf[1]==1)
              { /* k is a prime */
                for(i=2;i<=k-2;i++) ord[powtmp[i]]=k;
              }
            else if(r==1 && powf[1]==2)
              { /* k is the square of a prime p1 */
                p1=(unsigned short int)pf[1];
                for(i=2;i<=k-2;i++) ord[powtmp[i]]=k;
                for(i=p1;i<k;i+=p1) ord[powtmp[i]]=p1;  /* exceptions */
              }
            else if(r==2 && powf[1]==1 && powf[2]==1)
              { /* k is product of 2 distinct primes p1, p2 */
                p1=(unsigned short int)pf[1];
                p2=(unsigned short int)pf[2];
                for(i=2;i<=k-2;i++) ord[powtmp[i]]=k;
                for(i=p1;i<k;i+=p1) ord[powtmp[i]]=p2;  /* exceptions */
                for(i=p2;i<k;i+=p2) ord[powtmp[i]]=p1;  /* exceptions */
              }
        /* k not included above: 12, 16, 18, 20, 24, 27, etc. */
      }
  }

for(i=2;i<=n;i++)   /* count # of elems of each order (2 to n) */
  { chr[ord[i]]++;
  }

/* if(chr[2]!=3) return; */  /* **1** skip if not particular group we want n=12 */

/* If requested, skip (i.e. do not process) the group if abelian */
if(g_skipabeliansw && chr[1]==0)
  { g_numabelskip++;  /* count abelian non-cyclic groups skipped */
    return;
  }

/* If requested, skip the group if non-abelian */
if(g_skipnonabelsw && chr[1]!=0 && !g_stopnonabelsw)
  { g_numnonabelskip++;  /* count non-abelian groups skipped */
    return;
  }

/* "Process" the group */
g_numgroups++;
if(chr[1]) g_numnonabel++;
      else g_numabelian++;
displsw=1;  /* init. switch: 1=display, 0=not */
previdnum=0;  /* if nonzero: id num of isomorphic previous group */
groupidnum=g_numgroups;  /* tentative group id num for save and display */
if(g_testisomsw)
  { /* Test whether this group is isomorphic to a previous saved group */
    newgrsw=1;  /* init. switch: new group */
    for(gi=g_firstsav;gi<g_nsavedgr;gi++)
      { getgroup(n,&previdnum,s,chrs,ords,g_savedgr[gi]);  /* Get saved group
                      into s,chrs,ords, as group G1. Current group is G2. */
        for(i=1;i<=n;i++)  /* Compare group "characteristics" */
          { if(chr[i]!=chrs[i]) { i=0; break; }  /* (use i as switch) */
          }
        if(i==0) continue;  /* not isom., try next saved group */
        if(g_debug>=5 && g_displayisomsw)
          { /* This msg is to make it clear that testisom() trace is for
               NEXT displayed group (groupidnum). */
            printf("\ntestisom() for isomorphism between %d. and %d.:\n",
                   previdnum,groupidnum);
          }
        if(testisom(n,s,t,ords,ord))
          { newgrsw=0;  /* isomorphic: not a new group */
            if(chr[1]==0) g_numisomabelian++;
                     else g_numisomnonabel++;
            if(!g_displayisomsw) displsw=0;
            break;  /* exit loop if isom. */
          }
      }
    if(newgrsw)  /* new group: save it */
      { previdnum=0;  /* not isom. to a previous group */
        if(!g_displayisomsw) groupidnum=g_nsavedgr-g_firstsav+1;
                                  /* assign id num (must be nonzero) */
              /* Note: Normally g_displayisomsw is 0, and only new groups
                 are displayed, with id num = x =
                 (index in save array) - g_firstsav + 1.
                 But if g_displayisomsw is 1, all processed groups are
                 displayed, and the id num = g_numgroups, which may be > x. */
        if(g_nsavedgr>=MAXSAV)
          { printf("**** Table of saved groups is full. Group not saved. "
                   "Max is %d.\n",MAXSAV);
          }
        else
          { savgroup(n,groupidnum,t,chr,ord,&savptr);
            g_savedgr[g_nsavedgr]=savptr;  /* save memory pointer in table */
            g_nsavedgr++;
          }
      }
  }

if(displsw && !g_nodisplaysw)
  {
    printf("\n%5d. ",groupidnum);  /* display the mult. table */
    for(i=1;i<=n;i++)
      { for(j=1;j<=n;j++) printf(" %3hu",t[i][j]);
        printf("\n       ");
      }
    if(chr[1]) printf("Non-abelian\n");
          else printf("Abelian\n");
    printf("       # elems of order 2,3,..n:");
    for(i=2;i<=n;i++) printf(" %3hu",chr[i]); printf("\n");
    printf("       order of element 2,3,..n:");
    for(i=2;i<=n;i++) printf(" %3hu",ord[i]); printf("\n");
    if(previdnum) printf("      Isomorphic to previous group %d.\n",previdnum);
  }

/* Special handling for the group. */
if(g_debug==12)
  { if(groupidnum==1) savgroup(n,groupidnum,t,chr,ord,&g_savgrptr);  /* save it */
    if(groupidnum==2) chkisom(t,n,chr,ord,g_savgrptr);  /* isomorphic? */
  }

if(g_stopnonabelsw && chr[1]!=0)
  { printf("** Exiting because first non-Abelian group found\n");
    printf("   Number of groups found of order %hu: %d\n",n,g_numgroupsf);
    exit(2);
  }

/* if(chr[2]==3) exit(8); */  /* **1** stop if particular group found */

}  /* end of found */

/* ---------------------------------------------------------------------- */

static void savgroup(unsigned short int n, int idnum,
     unsigned short int t[MAX1][MAX1], unsigned short int chr[MAX1],
     unsigned short int ord[MAX1], void **psavptr)
/* This routine saves a found group (arrays t,chr,ord) into a memory
   area (allocated by this routine) and sets *psavptr to the pointer
   to the area. chr[],ord[] contain info about the group: see found() routine.
   Input args: n, idnum, t, chr, ord
   Output args: savptr (=*psavptr)
*/
{
struct savedgroupstruct *area;
unsigned long int totlen;

totlen=sizeof(struct savedgroupstruct);  /* or: sizeof(*area) */
area=malloc(totlen);
if(area==NULL)
  { printf("** Error in savgroup(): unable to get memory\n");
    exit(4);
  }
area->n=n;
area->idnum=idnum;
memcpy(area->t,t,sizeof(area->t));
memcpy(area->chr,chr,sizeof(area->chr));
memcpy(area->ord,ord,sizeof(area->ord));
if(g_debug==12) printf("Group info saved, %lu bytes\n",totlen);
/* printf("* in savgroup(): totlen=%lu  t size=%d\n",totlen,sizeof(area->t)); */
*psavptr=area;
}  /* end of savgroup */

/* ---------------------------------------------------------------------- */

static void getgroup(unsigned short int n, int *pidnum,
     unsigned short int t[MAX1][MAX1], unsigned short int chr[MAX1],
     unsigned short int ord[MAX1], void *savptr)
/* This routine gets a saved group (sets arrays t,chr,ord) from a memory
   area. chr[],ord[] contain info about the group: see found() routine.
   Input args: n, savptr
   Output args: idnum (=*pidnum, group id num), t, chr, ord
*/
{
struct savedgroupstruct *area;

area=savptr;
if(area==NULL)
  { printf("** Error in getgroup(): area pointer is NULL\n");
    exit(5);
  }
if(area->n != n)
  { printf("** Error in getgroup(): order %hu of saved group is not %hu\n",
           area->n,n);
    exit(5);
  }
    /* Note: can't use sizeof(t) because, surprisingly, that gives
       value 4, not the size of the array. I guess because t is an arg. */
*pidnum=area->idnum;
memcpy(t,area->t,sizeof(area->t));
memcpy(chr,area->chr,sizeof(area->chr));
memcpy(ord,area->ord,sizeof(area->ord));
}  /* end of getgroup */

/* ---------------------------------------------------------------------- */

static int chkisom(unsigned short int t[MAX1][MAX1],
     unsigned short int n, unsigned short int chrt[MAX1],
     unsigned short int ordt[MAX1], void *savptr)
/* This routine tests whether the group passed to it (table t, info chrt,ordt)
   is isomorphic to a previously saved group (area pointed to by savptr).
   Return value is 1 if isomorphic, 0 if not.

   **NOTE: This routine is not general. It is specific to
       a particular n (group order) and groups. See g_debug=12 coding.
       It is separate from the general routines (testisom(), etc.).
       (This routine could be removed, now that testisom() exists.)

   This version: for n=21, to compare the 2 non-Abelian groups found.
     Specify options g_debug 12, Skip abelian: y, Test for isomorphisms: n.
     May be specific to version 3 of groups.c .
*/
{
unsigned short int i,j;
int idnum;
unsigned short int s[MAX1][MAX1], chrs[MAX1], ords[MAX1];  /* saved group */
/* unsigned short int f[21+1]=
   {0,1,3,5,7,2,4,6,  9,8, 13,12, 17,16, 21,20, 11,10, 15,14, 19,18 }; */
                           /* Both f's given here are isomorphisms */
unsigned short int f[21+1]=
   {0,1,2,3,4,5,6,7,  9,8, 11,10, 13,12, 15,14, 17,16, 19,18, 21,20 };
     /* f[] defines the function: G1 (s) --> G2 (t) that is
        a possible isomorphism between groups G1 and G2.
        In each group, elems 2 to 7 are a to a**6; elems 8,9, 10,11, .. 20,21
        are b,b**2, ab,ab**2, .. (a**6)b,(a**6)(b**2). (See how the table is
        initialized in main().)  a has order 7, b has order 3.
        In G1, ba = (a**2)b.  In G2, ba = (a**4)b.
        The key to the isomorphism is that, if f is an isomorphism,
        f(ba) = f((a**2)b) = f(a**2)f(b) = ((f(a))**2)f(b) = (x**2)y,
        where x=f(a) and y=f(b). Also f(ba) = f(b)f(a) = yx.
        Therefore we need x and y such that (x**2)y = yx in G2.
        Also, since isomorphism preserves the order of each element,
        x must be a power of a and y must be a power of b.

        If we compute (b**2)a in G2 (according to the rule ba=(a**4)b),
        we get (a**2)(b**2), therefore we can try x=a, y=b**2, i.e.
        f(a) = a, f(b) = b**2. This gives f for all the other elems of G1.
        (The resulting f is an isomorphism, the same one found by testisom().)

        2nd isomorphism:
        If we compute (b**2)(a**2) in G2 (according to the rule ba=(a**4)b),
        we get (a**4)(b**2), therefore we can also try x=a**2, y=b**2, i.e.
        f(a) = a**2, f(b) = b**2. This gives f for all the other elems of G1.
        (The resulting f, commented out above, is also an isomorphism.)

        ---> This shows that, even when 2 groups are defined by generating
             elements and different defining relations, they may be
             isomorphic, and the isomrphism may not be obvious!
             Also, 2 group multiplication tables may be isomorphic by
             more than one isomorphism function.
     */

if(n!=21)
  { printf("** Wrong n in chkisom(). Expecting n=21.\n");
    return 0;
  }

getgroup(n,&idnum,s,chrs,ords,savptr);  /* get saved group (sets s,chrs,ords) */

/* First check that the group info is the same: Abelian/non-Abelian, etc. */
for(i=1;i<=n;i++)
  { if(chrs[i]!=chrt[i])
      { printf("** In chkisom(): groups NOT isomorphic (different info)\n");
        return 0;
      }
  }

/* Test whether the function f:G1-->G2 is an isomorphism.
   We assume always f(1)=1, therefore we don't need to check products
   involving the identity element. */
for(i=2;i<=n;i++) for(j=2;j<=n;j++)   /* check that f(ij)=f(i)f(j) */
  { if(f[s[i][j]]!=t[f[i]][f[j]])
      { printf("** In chkisom(): groups NOT isomorphic. Failed i=%hu j=%hu\n",
               i,j);
        return 0;
      }
  }

printf("** In chkisom(): the 2 groups are isomorphic, by the function f =\n");
for(i=1;i<=n;i++) printf(" %2hu",f[i]);
printf("\n");

/* Also test by the general function testisom() */
if(g_debug>=6) g_debug=5;  /* disable debug tracing in isomok() */
testisom(n,s,t,ords,ordt);

return 1;
}  /* end of chkisom */

/* ---------------------------------------------------------------------- */

static void factor(unsigned int n, unsigned int pf[], unsigned int pow[],
              unsigned int *r)
/* Factor n into a product of powers of r distinct primes, with the
   primes in increasing order.  It sets r.  n = product of pf[i]**pow[i]
   (i=1 to r). pow[i]>0. If n<2, r is set to 1.
   It is assumed that the arrays pf[] and pow[] provided by the caller
   are large enough for n.
*/
{
int ksw;
unsigned int k,div,rest;

if(n<2) { *r=0; return; }
for(k=0,ksw=0,div=2,rest=n;div*div<=rest;)
  { if(rest%div==0)
      { if(ksw==0) { k++; pf[k]=div; ksw=1; pow[k]=0; }
        pow[k]++;
        rest/=div;
      }
    else
      { if(div==2) div=1;  /* adjust for first div being even */
        div+=2;  /* next trial divisor */
        ksw=0;  /* next divisor (or rest) will be new prime */
      }
  }
if(rest>1)
  { /* rest is a prime. */
    if(ksw && rest==pf[k]) pow[k]++;
    else { k++; pf[k]=rest; pow[k]=1; }
  }
*r=k;
}  /* end of factor */

/* ---------------------------------------------------------------------- */

static int testisom(unsigned short int n,
          unsigned short int s[MAX1][MAX1], unsigned short int t[MAX1][MAX1],
          unsigned short int ords[MAX1], unsigned short int ordt[MAX1])
/* This routine tests whether group G1 (s,ords) is isomorphic to group G2
   (t,ordt).  s and t are the multiplication tables.  ords[i] and ordt[i]
   (i=1 to n) are the orders of element i in G1 and G2 respectively,
   provided by the caller.  The caller should compare the info arrays
   (chrs[], chrt[]) before calling this routine: if they are different,
   then G1 and G2 are not isomorphic.  This routine returns 1 if
   isomoorphic, 0 if not.  Routines nextisom() (recursive) and isomok()
   are used.
*/
{
unsigned short int i;
int rc;
unsigned short int f[MAX1];  /* The possible isomorphism function f
                     from G1 to G2.  f[i] (i=1 to n) is element number
                     in G2 of f(element number i in G1), or is 0 if
                     unknown so far. */

g_numtestisomcalls++;
f[1]=1;  /* initilaize f */
for(i=2;i<=n;i++) f[i]=0;

g_isomnestlevel=0;
rc=nextisom(n,f,s,t,ords,ordt);   /* do it! */

if(g_isomnestlevel!=0) printf("*** Warning: g_isomnestlevel = %d should be 0 at end\n",
                      g_isomnestlevel);

if(g_debug>=5)   /* (This is mainly for g_debug=12 or similar) */
  { if(rc)
      { printf("** testisom() found the 2 groups to be isomorphic by f =\n");
        for(i=1;i<=n;i++) printf(" %2hu",f[i]);
        printf("\n");
      }
    else printf("** testisom() found the 2 groups NOT isomorphic\n");
  }

return rc;
}  /* end of testisom */

/* ---------------------------------------------------------------------- */

static int nextisom(unsigned short int n, unsigned short int f[MAX1],
          unsigned short int s[MAX1][MAX1], unsigned short int t[MAX1][MAX1],
          unsigned short int ords[MAX1], unsigned short int ordt[MAX1])
/* This is a recursive function that tries all the possibilities for
   extending (i.e. completing the definition of) function f: G1 --> G2
   to am isomorphism. If f cannot be extended to an isomorphism, it returns
   0 (and f[] is not changed). If f can be extended, it returns 1 (and f
   is completely set), meaning that groups G1 and G2 are isomorphic.

   See comments in isomok() below for description of args n, f, s, t.
   See comments in testisom() above for description of args ords, ordt.

   Summary of method:
   We use the fact that an isomorphism preserves the order of an element,
   and, in fact, f(a**k)=(f(a))**k. We find the first undefined f[i]
   for an element i of highest order in G1. Then the choices for f[i] are
   the elements of G2 that have order = order of (element i in G1) and
   are not already an image of f. For each of these choices, we set
   f[i] and call isomok(), which determines whether f can be extended to
   an isomorphism. Note that isomok() may change f, therefore we must
   save a copy of the original f, to use for testing the next choice.
   We assume the f passed to us is a valid isomorphism so far.
   We assume the caller of nextisom() has already set f[1]=1.
   The maximum call nest depth for nextisom() is about n, and is usually
   considerably less than n, since isomok() (and this routine) may
   fill in several f[i].
*/
{
unsigned short int i,j,ii,fndi,fndord,fndord1,powi,powj,fofpowi;
unsigned short int savef[MAX1];  /* for saving original f[] */
int oksw,cntfset,savecntfset;

g_numnextisomcalls++;
g_isomnestlevel++;  /* decrement before return */
if(g_isomnestlevel>g_isomhinestlevel) g_isomhinestlevel=g_isomnestlevel;

/* Find first i such that f[i] is not defined and elem i has highest
   order in G1. Also count the number of elements defined in f. */
for(fndi=0,cntfset=1,i=2;i<=n;i++)
  { if(f[i]) { cntfset++; continue; }
    if(fndi)
      { if(ords[i]>fndord) { fndi=i; fndord=ords[i]; }
      }
    else { fndi=i; fndord=ords[i]; }
  }
if(fndi==0) { g_isomnestlevel--; return 1; }
         /* f is all defined, and so it is an isomorphism */

if(g_debug>=5)
  { printf("nextisom() called: fndi=%hu fndord=%hu f =",fndi,fndord);
    for(i=1;i<=n;i++) printf(" %2hu",f[i]);
    printf("\n");
  }

for(i=2;i<=n;i++) savef[i]=f[i];  /* save original f */
savecntfset=cntfset;  /* also save the set count */

/* Check each possible choice for f[fndi]. Each choice (an element of G2)
   must have order fndord and not already be an image of f. */
for(j=2;j<=n;j++)   /* j is an elem of G2 */
  { if(ordt[j]!=fndord) continue;  /* skip if different order */
    for(i=2;i<=n;i++)
      { if(j==f[i]) { i=0; break; }  /* skip if already an image of f */
      }
    if(i==0) continue;
    f[fndi]=j;  /* set f for this choice */
    cntfset++;
    if(g_debug>=5)
      { printf("nextisom(): trying f[%hu]=%hu\n",fndi,j);
        /* To get debug tracing in isomok(), set g_debug=6 or higher here */
      }
    /* At this point we could just call isomok(), which would fill in
       f[] for the powers of element fndi, but it's faster to fill
       them in here, before calling isomok(). */
    oksw=1;
    if(fndord>2)
      { fndord1=fndord-1;
        for(powi=fndi,powj=j,i=2;i<=fndord1;i++)
          { powi=s[powi][fndi];  /* get powi = (elem fndi)**i in G1 */
            powj=t[powj][j];  /* get powj = (elem j)**i in G2 */
            /* Now we want to set f[powi]=powj */
            /* Note: We don't need to specifically set f(inverse of powi)
               or f(inverse of fndi), since the inverse of any power of
               fndi is also a power of fndi. */
            fofpowi=f[powi];
            if(fofpowi)  /* already set */
              { if(fofpowi!=powj) { oksw=0; break; }   /* not isom. */
              }
            else
              { /* Check that powj is not already an image of f.
                   (We know powj is not 1.) */
                for(ii=2;ii<=n;ii++)
                  { if(powj==f[ii]) { oksw=0; break; }
                  }
                if(!oksw) break;
                f[powi]=powj;  /* Set f. Later call to isomok() will check f. */
                cntfset++;
                if(g_debug>=5) printf("  nextisom() setting f[%hu]=%hu (power)\n",
                                      powi,powj);
              }
          }
      }
    if(oksw)
      { /* Call routine to check f, and possibly define more of it.
           Note: isomok() does not update our cntfset even if it defines more
           of f. That's ok, since we have finished with cntfset now. */
        if(isomok(n,f,s,t,cntfset))
          { /* Ok so far. Do recursive call, to continue extending f */
            if(nextisom(n,f,s,t,ords,ordt)) { g_isomnestlevel--; return 1; }
                           /* isom. found! */
          }
      }
    /* If isomorphism not found yet, continue to next choice. */
    /* Restore original f and cntfset before doing next choice or returning 0 */
    for(i=2;i<=n;i++) f[i]=savef[i];
    cntfset=savecntfset;
  }

if(g_debug>=5) printf("nextisom() returning 0\n");
g_isomnestlevel--;
return 0;   /* no extension of f to an isomorphism is possible */
}  /* end of nextisom */

/* ---------------------------------------------------------------------- */

static int isomok(unsigned short int n, unsigned short int f[MAX1],
          unsigned short int s[MAX1][MAX1], unsigned short int t[MAX1][MAX1],
          int cntfset)
/* This routine tests whether the isomorphism function f (which may be
   only partially defined, i.e. some values may be 0) satisfies the
   isomorphism requirement f(xy)=f(x)f(y) so far. It returns 0 if no,
   1 if yes. In either case, it may define (i.e. set f[i] nonzero)
   additional values for f. Table s is the multiplication table of group G1.
   Table t is the multiplication table of group G2.  f is a function
   from G1 to G2.  f[i] (i=1 to n, where n is the order of the groups)
   is the element # (in G2) that f maps element # i (of G1) to, or is
   0 if the image of element i is not defined yet. f[1] is always 1,
   since element 1 is always the identity element.
   Arg cntfset is the number of values defined so far in the f passed
   to us. This routine uses this count to know when f is complete,
   but a full check is always done after setting f.
   CAUTION: Before setting an f[i] to j, make sure j is not already an
   image of f, since f must be 1-to-1.
   ALSO: When setting f[i] to j, normally also set f[inverse of i] to
   the inverse of j, since an isomorphism must preserve inverses.
*/
{
unsigned short int i,j,ii,fofi,fofj,sprod,tprod,sprod1,tprod1;
int setsw;  /* nonzero means f definition(s) set by a loop iteration */

g_numisomokcalls++;
if(g_debug>=6)
  { printf("isomok() called, cntfset=%d  f =\n",cntfset);
    for(i=1;i<=n;i++) printf(" %2hu",f[i]);
    printf("\n");
  }
for(;;)  /* this loop allows redo if element(s) of f were set */
  { for(setsw=0,i=2;i<=n;i++)  /* no need to involve the identity element */
      { fofi=f[i];
        if(fofi==0) continue;  /* skip this i if f(i) not defined yet */
        for(j=2;j<=n;j++)
          { fofj=f[j];
            if(fofj==0) continue;  /* skip if f(j) not defined yet */
            sprod=s[i][j];  /* product of elements in G1 */
            tprod=t[fofi][fofj];  /* product of elements in G2 */
            if(g_debug>=6) printf("  isomok() checking i=%hu j=%hu "
                     "f[i]=%hu f[j]=%hu sprod=%hu tprod=%hu\n",
                     i,j,f[i],f[j],sprod,tprod);
            if(f[sprod]==0)
              { /* We want to set f[sprod]=tprod, but first check that
                   tprod is not already an image of f. */
                for(ii=1;ii<=n;ii++)
                  { if(tprod==f[ii]) { ii=0; break; }  /* (use ii as a switch) */
                  }
                if(ii==0) return 0;  /* f is not an isomorphism */
                f[sprod]=tprod;
                cntfset++;
                if(g_debug>=6) printf("     set f[%hu]=%hu\n",sprod,tprod);
                setsw=1;
                if(cntfset==n) break;  /* exit this checking loop if f complete */
                /* Also set f(inverse of sprod), if possible.
                   Note: neither of sprod,tprod is 1. */
                for(sprod1=2;sprod1<=n;sprod1++) if(s[sprod][sprod1]==1) break;
                                             /* get sprod1 = inv. of sprod */
                if(sprod1!=sprod)
                  { for(tprod1=2;tprod1<=n;tprod1++) if(t[tprod][tprod1]==1) break;
                                             /* get tprod1 = inv. of tprod */
                    if(f[sprod1]==0)
                      { for(ii=1;ii<=n;ii++)
                          { if(tprod1==f[ii]) { ii=0; break; }
                          }
                        if(ii==0) return 0;
                        f[sprod1]=tprod1;
                        cntfset++;
                        if(g_debug>=6) printf("     set f[%hu]=%hu (inv)\n",
                                          sprod1,tprod1);
                        if(cntfset==n) break;  /* exit this checking */
                      }
                    else
                      { if(f[sprod1]!=tprod1) return 0;  /* f is not an isom. */
                      }
                  }
              }
            else
              { if(f[sprod]!=tprod) return 0;  /* f is not an isomorphism */
              }
          }  /* end of j loop */
        if(cntfset==n && setsw) break;  /* exit this checking loop if f complete */
      }  /* end of i loop */
    if(setsw==0) break;  /* redo checking if we changed f */
  }
if(g_debug>=6) printf("  isomok() returning 1\n");
return 1;  /* f is ok so far */
}  /* end of isomok */
